/*------------------------------------------------------------------------

  File: tt-file.p

  Description: Convert a temp table into a file.

  Input Parameters:
    temp table handle
    filename
    option list

  Output Parameters:
    <none>

  Author: Jeff Pilant

  Usage and documentation: see tt-file.html

  ------------------------------------------------------------------------
  Contributers:
    Who  Full name         <email>
    JTP  Jeff Pilant       <Jeff.Pilant@us.abb.com>
    PT   Patrick Tingen    <p.tingen@vcd.nl>

  ---------------------------------------------------------------------------
  Revision History:

  Ver  Who  Date         Description
  ---- ---- ------------ --------------------------------------------------
  1.0  JTP  06-Mar-2002  Created (TXT, XLS)
       JTP  06-Mar-2002  Added (CSV), prompt for filename
       JTP  ??-???-2002  Added (BRS)
       JTP  11-Apr-2002  Added AutoFit to Excel output
       JTP  23-May-2002  Fixed getExcelColumn()
       JTP  08-May-2002  Changed default column width for excel to a
                         function
                         Added Title Row code to Excel output
       JTP  13-Sep-2002  Added code to fix Title Row to always be row 1
       JTP  10-Oct-2002  Added ability to do csv, txt, or xls files if
                         type not given
       JTP  16-Jul-2003  Added Sylk file type
  1.1  JTP  08-Sep-2003  Fixup Sylk date format for dates before 1/1/1900
                         Added df file type [From code found at:
                           http://www.v9stuff.com/dynexport.htm
                           thanks to Tony Lavinio and Peter van Dam]
       PT   22-Oct-2003  Added, "character" to substring and length
                         functions to support multi-byte codepages.
                         [JTP: Adds compatibility for other Code Pages]
  1.2  PT   31-Oct-2003  Changed parameter cFileType to pcOptions to provide
                         more options for the requested behaviour.
                         This is backward compatible.
                         Added options: FileType, FieldList, SkipList,
                                        ExcelAlert and ExcelVisible
       JTP  01-Nov-2003  Folded in PT's changes and rewrote the column
                         autofit portion of the excel code.
       PT   02-Dec-2003  Added options: Grid, GridCharHor, GridCharVer,
                                        GridCharCross, GridInterval, GridField
                         Added options: Append, Labels, FirstOfList, Title
                         Added HTML documentation.
                         Added XML mode.
                         Moved code and vars to internal procedures.
       PT   12-Jan-2004  Use the name of the field as the XML tag, not the label
                         since the label is more likely to hold undesired chars.
                         Added some comments to procedures and functions.
       JTP  21-Jan-2004  Fix Sylk bug quoting semicolons
       PT   09-Feb-2004  Added support for writing array fields to XML
  1.3  JTP  12-Feb-2004  Rewrote getXMLFriendlyString() to better conform
                         to http://www.w3.org/TR/2004/REC-xml-20040204/
                         for XML Names.
                         Reordered DecodeOptionString() case to match
                         documentation sequence.
                         Added XML filetype to GetOutputFile().
                         Added windows check to GetOutputFile() for blank
                         filename. [SYSTEM-DIALOG GET-FILE is windows only]
                         Added windows check for Excel output.
                         Fine tuned GetOutputFile()'s use of SYSTEM-DIALOG
                         Added SubTotal and SubBreak options.
                         Extracted title and query code from TXT, CSV, XLS
                         and SLK
                         Added subtotal break detection code to TXT, CSV,
                         XLS and SLK
                         Removed Append option from DF files, as it does
                         not make sense.
                         Corrected Label printing for XLS and CSV.
                         Added Append option to XLS.  It forces ExcelAlert
                         to False (don't ask on overwrite).
       PT   18-Feb-2004  Moved code to function getFieldName.
                         Added support for counting date fields.
                         TXT output now respects leading spaces in format.
                         Remove fields from SubTotal when they are not
                         included in the list of fields to be shown.
                         Moved code to determine fieldname to getFieldName.
                         Added options: Where, Sort
                         Updated HTML documentation up to this point.
       PT   23-Feb-2004  Added options: BrowseWidth, BrowseHeight, Table,
                         BrowseTooltip, Multiple, ColumnSortList
                         Added save-file dialog for non-win32 client.
                         If no field is specified, the value of the first
                         field in the buffer will be returned (brs).
                         Accept a buffer handle instead of a tt-handle.
                         Accept table name instead of a tt-handle or buffer handle.
                         Exclude a field from tooltip if it is not required (brs)
                         Fixed bug when exporting nodes with ? value (xml)
                         Allowed for column-sorting by clicking on it (brs)
                         Allowed for multiple rows (brs)
                         Warn on unsupported options.

------------------------------------------------------------------------*/
define input parameter phSource   as handle    no-undo.
define input parameter pcFileName as character no-undo.
define input parameter pcOptions  as character no-undo.

/* Excel Constants */
&scoped-define xlExcel9795         43
&scoped-define xlWBATWorksheet    -4167
&scoped-define xlWorkbookNormal   -4143
&scoped-define xlRight            -4152

/* Sylk Constants */
&scoped-define SylkFormat-number  "N":U
&scoped-define SylkFormat-string  "S":U
&scoped-define SylkFormat-date    "D":U
&scoped-define SylkFormat-time    "T":U
&scoped-define SylkFormat-logical "L":U
&scoped-define sylk-xls-hm        "HH:MM":U
&scoped-define sylk-xls-hm-a      "HH:MM AM":U
&scoped-define sylk-xls-hms       "HH:MM:SS":U
&scoped-define sylk-xls-hms-a     "HH:MM:SS AM":U
&scoped-define sylk-xls-mdy       "MM/DD/YY":U
&scoped-define sylk-xls-dmy       "DD-MMM-YY":U
&scoped-define sylk-xls-md-y      "Month DD, YYYY":U
&scoped-define sylk-xls-ymd       "YYMMDD":U
&scoped-define sylk-xls-cymd      "YYYYMMDD":U
&scoped-define sylk-xls-title     "TITLE":U
&scoped-define sylk-xls-body      "BODY":U
&scoped-define sylk-fil-picture   "P":U
&scoped-define sylk-fil-font      "M":U
&scoped-define sylk-fil-c-wid     "W":U
&scoped-define sylk-fil-c-fmt     "C":U
&scoped-define sylk-fil-c-title   "T":U

/* Browse constants */
&scoped-define BrowseMaximumWidth     320
&scoped-define BrowseMinimumWidth     38
&scoped-define BrowseDefaultWidth     90

/* Ascii constants */
&scoped-define lf                 chr(10)
&scoped-define cr                 chr(13)
&scoped-define qt                 chr(34)
&scoped-define mk                 chr(1)
&scoped-define sComma             ",":U
&scoped-define sColon             ":":U
&scoped-define sChar              "character":U

/* Include files */
{tt-file.i}

/* General variables */
define variable cTableName           as character no-undo. /* for db-usage */
define variable nf                   as integer   no-undo.
define variable nx                   as integer   no-undo.
define variable wf                   as integer   no-undo.
define variable nc                   as integer   no-undo.
define variable btt                  as handle    no-undo.
define variable hwf                  as handle    no-undo.
define variable hQuery               as handle    no-undo.
define variable hField               as handle    no-undo.
define variable cWhere               as character no-undo. /* Where clause on the query */
define variable cSort                as character no-undo. /* Sort on the query */
define variable cBreakBy             as character no-undo. /* Break-by info (based on sort field) */
define variable cFileType            as character no-undo.
define variable cOutputFile          as character no-undo.
define variable cBrowseTitle         as character no-undo initial "tt-browse":U.
define variable cFieldList           as character no-undo initial "*":U. /* by default, include all fields */
define variable cSkipList            as character no-undo initial "".    /* by default, exclude no fields  */
define variable lFirstField          as logical   no-undo.               /* var to simulate a first-of     */
define variable lAppend              as logical   no-undo.               /* append to a file or not        */
define variable lPrintLabels         as logical   no-undo initial true.  /* show header with labels        */
define variable lDelBuffer           as logical   no-undo initial false. /* delete buffer when done?       */
define variable cSubBreak            as character no-undo initial "".
define variable iSubBreak            as integer   no-undo initial 0.
define variable cSubTotal            as character no-undo initial "".
define variable str                  as character no-undo.
define variable tag                  as character no-undo.
define variable lEof                 as logical   no-undo.
define variable lEol                 as logical   no-undo.
define variable iPrnField            as integer   no-undo.
define variable iNumField            as integer   no-undo.
define variable iCurField            as integer   no-undo.
define variable iSeqField            as integer   no-undo.
define variable iNumIndex            as integer   no-undo.
define variable iCurIndex            as integer   no-undo.
define variable cFieldStr            as character no-undo.
define variable cFieldBuf            as character no-undo.
define variable cFieldName           as character no-undo.
define variable cFieldLabel          as character no-undo.
define variable cFieldFormat         as character no-undo.
define variable cBufComp             as character no-undo.
define variable tth                  as handle    no-undo.
define variable hTmpBuf              as handle    no-undo. /* buffer used for change detection */
define variable iRecordCounter       as integer   no-undo.
define variable brk-new              as integer   no-undo.
define variable brk-old              as integer   no-undo.
define variable brk-cur              as integer   no-undo.

/* TXT variables */
define variable cFirstOfList         as character no-undo.
define variable cGridCharHor         as character no-undo initial "-":U.
define variable cGridCharVer         as character no-undo initial "|":U.
define variable cGridCharCross       as character no-undo initial "+":U.
define variable iGridInterval        as integer   no-undo initial 1.
define variable cHorGridFieldName    as character no-undo initial "".
define variable lHorGrid             as logical   no-undo initial no.
define variable lVerGrid             as logical   no-undo initial no.
define variable cThisRec             as character no-undo.
define variable cPrevRec             as character no-undo.

/* Excel variables */
define variable lExcelAlert          as logical   no-undo initial true. /* by default, complain if a file exists */
define variable lExcelVisible        as logical   no-undo initial true. /* by default, show excel */

/* BRS variables */
define variable cReturnString        as character no-undo.
define variable cReturnList          as character no-undo.
define variable iBrowseWidth         as integer   no-undo initial ?.  /* Actual width of the browse */
define variable iBrowseHeight        as integer   no-undo initial 15. /* Nr of rows to show on the browse */
define variable lBrowseTooltip       as logical   no-undo initial true. /* To show or not to show tooltips */
define variable cSortMethod          as character no-undo.
define variable cSortColumn          as character no-undo.
define variable cSortColumnPrev      as character no-undo.
define variable cColumnSortList      as character no-undo initial '*'. /* List of fields that can be clicked */
define variable lMultiple            as logical   no-undo initial false.

/* XML variables */
define variable cKeyFieldList        as character no-undo.

/* Structure for holding Sylk parts */
define temp-table tt-sylk no-undo
  field typ as character
  field seq as integer
  field val as character
  field fmt as character
  index pu-one is primary unique typ seq.

/* Structure for holding subtotals */
define temp-table tt-subtot no-undo
  field lvl     as integer   /* level of break      */
  field bk-name as character /* field to break on   */
  field st-name as character /* field to total on   */
  field st-pos  as integer   /* subtotal position   */
  field st-fmt  as character /* subtotal format     */
  field tot-typ as integer   /* Field type to total */
  field tot-int as integer   /* Total for type integer */
  field tot-dec as decimal   /* Total for type decimal */
  field tot-log as integer   /* Total for type logical */
  field tot-cnt as integer   /* Total for type character and date */
  index pu-one is primary unique lvl st-name bk-name
  index si-two   st-name
  index si-three bk-name
  index si-four  lvl st-pos.
&scoped-define tot-as-int 1
&scoped-define tot-as-dec 2
&scoped-define tot-as-log 3
&scoped-define tot-as-cnt 4

define temp-table tt-labels
  field bk-name as character /* field to break on   */
  field st-lab  as character /* field label         */
  field st-brk  as character /* value of break      */
  index pu-one is primary unique bk-name.

/* Structure & query for the browse */
define temp-table q-help no-undo
  field a as integer.

/*------------------------------------------------------------------------- *\
** function definitions
\*------------------------------------------------------------------------- */

/* Name : getExcelColumn
 * Desc : Get Excel column reference from column number.
 */
function getExcelColumn returns character
  ( input pos as integer ) :
  define variable ifirst  as integer   no-undo.
  define variable isecond as integer   no-undo.
  define variable cols as character no-undo.
  assign ifirst  = integer(truncate((pos - 1) / 26, 0))
         isecond = pos - (26 * ifirst)
         cols    = chr(64 + isecond).
  if ifirst > 0 then
    cols = chr(64 + ifirst) + cols.
  return cols.
end function. /* getExcelColumn */


/* Name : getLabelWidth
 * Desc : Get width needed for column using label and format.
 */
function getLabelWidth returns integer
  ( input hnd as handle ) :
  return integer(max(length(cFieldName, {&sChar}), hnd:width-chars)).
end function. /* getColumnWidth */


/* Name : getColumnWidth
 * Desc : Get width needed for column using label and format.
 */
function getColumnWidth returns integer
  ( input hnd as handle ) :
  return integer(max(length(hnd:label, {&sChar}), hnd:width-chars)).
end function. /* getColumnWidth */


/* Name : getExcelColumnWidth
 * Desc : Default column width to set Excel column.
 */
function getExcelColumnWidth returns integer
  ( input hnd as handle ) :
  return integer(min(1.25 * max(length(hnd:label, {&sChar}), hnd:width-chars), 255)).
end function. /* getExcelColumnWidth */


/* Name : getValidFileName
 * Desc : Convert string to valid filename.
 */
function getValidFileName returns character
  ( input cPar as character ) :
  define variable cFileName as character no-undo.
  cFileName = cPar.

  cFileName = entry(num-entries(cFileName, "\":U), cFileName, "\":U).
  cFileName = entry(num-entries(cFileName, "/":U), cFileName, "/":U).

  cFileName = replace(cFileName, "\":U, "_":U).
  cFileName = replace(cFileName, "/":U, "_":U).
  cFileName = replace(cFileName, ":":U, "_":U).
  cFileName = replace(cFileName, "*":U, "_":U).
  cFileName = replace(cFileName, "?":U, "_":U).
  cFileName = replace(cFileName, '"':U, "_":U).
  cFileName = replace(cFileName, "<":U, "_":U).
  cFileName = replace(cFileName, ">":U, "_":U).
  cFileName = replace(cFileName, "|":U, "_":U).

  if cFileName = "" then cFileName = "[blank]":U.
  cFileName = trim(cFileName).
  cFileName = substring(cFileName, 1, 32, {&sChar}).
  return cFileName.
end function. /* getValidFileName */


/*
 * Name : getFieldName
 * Desc : return the name of the field, if iExt > 0 with the addition
 *        of the extent between square brackets.
 */
function getFieldName returns character (hField as handle, iExt as integer):
  return substitute("&1&2"
                   , hField:name
                   , ( if iExt = 0 then "" else substitute("[&1]":U, iExt))
                   ).
end function. /* getFieldName */


/* Name : SylkValue
 * Desc : Get the reference used for a part.
 */
function SylkValue returns character
  ( input p-typ as character,
    input p-fmt as character ) :
  find first tt-sylk where tt-sylk.typ = p-typ
                       and tt-sylk.fmt = p-fmt
                           no-lock no-error.
  if available(tt-sylk) then
    return p-typ + string(tt-sylk.seq).
  else
    return p-typ + "0":U.
end function. /* SylkValue */


/* Name : sylk-str
 * Desc : Double semicolons and strip trailing spaces for sylk strings
 */
function sylk-str returns character
  ( input p-val as character ) :
  p-val = replace(right-trim(p-val), ";":U, ";;":U).
  return p-val.
end function.


/* Name : FieldRequired
 * Desc : Determine if a given field should be displayed.
 */
function FieldRequired returns logical
  ( input hField as handle ):
  /* cFieldList holds a list of fields to use  (default all)  */
  /* cSkipList  holds a list of fields to skip (default none) */
  return (        can-do(cFieldList, hField:name)
          and not can-do(cSkipList,  hField:name)).
end function. /* FieldRequired */


/* Name : LogicalValue
 * Desc : Logical function for pre 9.1D versions.
 */
function LogicalValue returns logical
  ( input cString as character ):
  return ( cString begins "Y":U or cString begins "T":U).
end function. /* LogicalValue */


/* Name : getXMLFriendlyString
 * Desc : Filter out characters that XML dislikes in an element name.
 */
function getXMLFriendlyString returns character
  ( input cString as character ):
  define variable cnt as integer   no-undo.
  define variable tmp as character no-undo.
  define variable cValidXMLNameInit    as character no-undo
    initial "ABCDEFGHIJKLMNOPQRSTUVWXYZ_":U.
  define variable cValidXMLNameAll     as character no-undo
    initial "ABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789.-":U.

  do cnt = 1 to length(cString, {&sChar}):
    tmp = substring(cString, cnt, 1).
    if cnt = 1 and index(cValidXMLNameInit, tmp) = 0 then tmp = ?.
    if cnt > 1 and index(cValidXMLNameAll,  tmp) = 0 then tmp = ?.
    if tmp = ? then substring(cString, cnt, 1) = "_":U.
  end.
  return cString.
end function. /* getXMLFriendlyString */


/* Name : display_label
 * Desc : Show display form of label.
 */
function display_label returns character
  (input str as character) :
  return replace(str, "_":U, " ":U).
end function.


/*------------------------------------------------------------------------- *\
** main code block begin
\*------------------------------------------------------------------------- */

run DecodeOptionString ( input pcOptions ).
run GetOutputFile ( output cOutputFile ).

/* Some checks:
   - File Type can be blank only if a filename was prompted and the
     dialog was canceled.
   - File name cannot be blank unless filetype is 'browse'
*/
if cFileType = "" then return.
if cFileType <> "BRS" and cOutputFile = "" then return.

/* If no handle has been passed, but a tablename has been given,
** create a buffer for that table.
*/
if phSource = ? and cTableName <> "" then
do:
  create buffer phSource for table cTableName no-error.
  assign lDelBuffer = true.
  if not valid-handle(phSource) then
  do:
    run fatalError(substitute("Invalid table name '&1' passed to tt-file.p":U, cTableName )).
    return.
  end.
end.

/* Set btt handle to the appropriate handle */
case phSource:type:
  when "temp-table":u then btt = phSource:default-buffer-handle.
  when "buffer":u     then btt = phSource.
end case. /* Buffer */

nf  = btt:num-fields.
cTableName = phSource:name.

create temp-table tth.
tth:create-like(btt).
tth:temp-table-prepare("tt-file-test":U).

/* Create a dynamic buffer for change detection */
hTmpBuf = tth:default-buffer-handle.
hTmpBuf:buffer-create.

brk-old = num-entries(cSubBreak).

case cFileType:
  when tt-df     then run df_format.
  when tt-text   then run txt_format.
  &if "{&opsys}":U = "WIN32":U &then
  when tt-excel  then run xls_format.
  &endif
  when tt-csv    then run csv_format.
  when tt-browse then run brs_format.
  when tt-sylk   then run slk_format.
  when tt-xml    then run xml_format.
  otherwise           run fatalError("Call to tt-file.p with unknown file type [":U + cFileType + "]":U).
end case.

/* Clean up */
hTmpBuf:buffer-release().
hTmpBuf = ?.

delete object tth.
tth = ?.

if lDelBuffer then delete object phSource.
phSource = ?.

/* If a return-value is composed, return it to the calling program */
return cReturnString.

/*------------------------------------------------------------------------- *\
** main code block end
\*------------------------------------------------------------------------- */


/* Name : DecodeOptionString
 * Desc : Decode option string into variables.
 */
procedure DecodeOptionString:
  define input parameter cOptionList as character no-undo.

  define variable iOption        as integer    no-undo.
  define variable cnt            as integer no-undo.
  define variable ndx            as integer no-undo.
  /* vars to pull the option string apart */
  define variable cCurrentOption as character  no-undo.
  define variable cCurrentValue  as character  no-undo.
  /* vars to examine cSubTotal and cSubBreak */
  define variable iField         as integer   no-undo.
  define variable cTmpFieldList  as character no-undo.
  define variable cTmpFieldName  as character no-undo.

  /* Decode option string */
  do iOption = 1 to num-entries(cOptionList, {&mk}):
    cCurrentOption = entry(iOption, cOptionList, {&mk}).
    cCurrentValue  = "".
    if num-entries(cCurrentOption, {&sColon}) > 1 then
      cCurrentValue = substring(cCurrentOption, index(cCurrentOption, {&sColon}) + 1, -1 , {&sChar}).
    if num-entries(cCurrentOption, {&sColon}) = 1 then
      cFileType = cCurrentOption. /* old way of providing file type */
    else
      case entry(1, cCurrentOption, {&sColon}):
        when "FileType":U      then cFileType         = cCurrentValue.
        when "FieldList":U     then cFieldList        = cCurrentValue.
        when "SkipList":U      then cSkipList         = cCurrentValue.
        when "Append":U        then lAppend           = LogicalValue(cCurrentValue).
        when "ExcelAlert":U    then lExcelAlert       = LogicalValue(cCurrentValue).
        when "ExcelVisible":U  then lExcelVisible     = LogicalValue(cCurrentValue).
        when "Title":U         then cBrowseTitle      = cCurrentValue.
        when "ReturnList":U    then cReturnList       = cCurrentValue.
        when "Labels":U        then lPrintLabels      = LogicalValue(cCurrentValue).
        when "FirstOfList":U   then cFirstOfList      = cCurrentValue.
        when "Grid":U          then
        do:
          /* Removed by PT request:
          ** Default the grid components
          ** assign
          **   lHorGrid          = no
          **   lVerGrid          = no
          **   cGridCharHor      = "-":U
          **   cGridCharVer      = "|":U
          **   cGridCharCross    = "+":U
          **   iGridInterval     = 1
          **   cHorGridFieldName = "".
          */
          /* Check for horizontal and vertical markings */
          case cCurrentValue:
            when "hor":U then lHorGrid = yes.
            when "ver":U then lVerGrid = yes.
            otherwise if LogicalValue(cCurrentValue) then
              assign
                lHorGrid = yes
                lVerGrid = yes.
          end case.
        end.
        when "GridCharHor":U   then cGridCharHor      = cCurrentValue.
        when "GridCharVer":U   then cGridCharVer      = cCurrentValue.
        when "GridCharCross":U then cGridCharCross    = cCurrentValue.
        when "GridInterval":U  then assign
                                    cHorGridFieldName = ""
                                    iGridInterval     = integer(cCurrentValue) no-error.
        when "GridField":U     then assign
                                    cHorGridFieldName = cCurrentValue
                                    iGridInterval     = 0.
        when "AttributeList":U then cKeyFieldList     = cCurrentValue.
        when "SubTotal":U      then cSubtotal         = cCurrentValue.
          /* SubTotal:field[,field]*
           *   List of fields to sum
           *   If numeric, then sum #
           *   If logical, then count # true entries
           *   If character, then count # entries
           *   If date, then count # entries
           *   If otherwise, skip
           */
        when "SubBreak":U      then cSubBreak         = cCurrentValue.
          /* SubBreak:field[,field]*
           *   List of fields to break on
           *   Grand total will be generated if option used
           */
        when "Table":U          then cTableName        = cCurrentValue.
        when "Where":U          then cWhere            = cCurrentValue.
        when "Sort":U           then cSort             = cCurrentValue.
        when "BrowseWidth":U    then iBrowseWidth      = integer(cCurrentValue).
        when "BrowseHeight":U   then iBrowseHeight     = integer(cCurrentValue).
        when "BrowseTooltip":U  then lBrowseTooltip    = LogicalValue(cCurrentValue).
        when "ColumnSortList":U then cColumnSortList   = cCurrentValue.
        when "Multiple":U       then lMultiple         = LogicalValue(cCurrentValue).

        otherwise message substitute("Unknown option '&1' ignored.", cCurrentOption )
                          view-as alert-box information buttons ok.
      end case.
  end.

  /* Force Excel to overwrite if Append is on */
  if lAppend then lExcelAlert = false.

  /* If no sort is specified and breaks are defined, set the sort var */
  if cSort = "" and cSubBreak <> "" then cSort = cSubBreak.

  /* Correct where clause if it does not begin with 'where' */
  if cWhere <> "" and not cWhere begins 'where ':U then cWhere = 'where ':U + cWhere.

  /* Create a BY clause based on the sort field */
  if cSort <> "" then
  do:
    cTmpFieldList = "".
    do iField = 1 to num-entries(cSort):
      cTmpFieldList = substitute("&1 by &2":u, cTmpFieldList, entry(iField, cSort)).
    end. /* cSort */
    cBreakBy = cTmpFieldList.
  end.

  /* Remove fields from SubTotal list when they are not needed */
  cTmpFieldList = "".
  do iField = 1 to num-entries(cSubTotal):
    cTmpFieldName = entry(iField,cSubTotal).
    if         can-do(cFieldList, cTmpFieldName)
       and not can-do(cSkipList,  cTmpFieldName) then
      assign cTmpFieldList = substitute("&1,&2":U,cTmpFieldList,cTmpFieldName).
  end. /* SubTotal */
  cSubTotal = trim(cTmpFieldList, ",":U).

  /* commented out by PT, see my email
  **
  **   /* Remove fields from SubBreak list when they are not needed */
  **   cTmpFieldList = "".
  **   do iField = 1 to num-entries(cSubBreak):
  **     cTmpFieldName = entry(iField,cSubBreak).
  **     if         can-do(cFieldList, cTmpFieldName)
  **        and not can-do(cSkipList,  cTmpFieldName) then
  **       assign cTmpFieldList = substitute("&1,&2":U,cTmpFieldList,cTmpFieldName).
  **   end. /* SubTotal */
  **   cSubBreak = trim(cTmpFieldList, ",":U).
  */

  /* check interval */
  if iGridInterval lt 0 then iGridInterval = 0.

  /* Set up subtotaling */
  do cnt = 1 to num-entries(cSubBreak): /* Generate a total at each level */
    do ndx = 1 to num-entries(cSubtotal): /* Total each listed column */
      create tt-subtot.
      assign tt-subtot.lvl     = cnt
             tt-subtot.bk-name = entry(cnt, cSubBreak)
             tt-subtot.st-name = entry(ndx, cSubtotal)
             tt-subtot.tot-int = 0
             tt-subtot.tot-dec = 0.0
             tt-subtot.tot-log = 0
             tt-subtot.tot-cnt = 0.
    end.
    create tt-subtot.
    assign tt-subtot.lvl     = 0
           tt-subtot.bk-name = entry(cnt, cSubBreak)
           tt-subtot.st-name = ""
           tt-subtot.tot-int = 0
           tt-subtot.tot-dec = 0.0
           tt-subtot.tot-log = 0
           tt-subtot.tot-cnt = 0.
  end.
  if cSubTotal <> "" then
  do ndx = 1 to num-entries(cSubtotal): /* If subtotals are defind, do a grand total */
    create tt-subtot.
    assign tt-subtot.lvl     = 0
           tt-subtot.bk-name = ""
           tt-subtot.st-name = entry(ndx, cSubtotal)
           tt-subtot.tot-int = 0
           tt-subtot.tot-dec = 0.0
           tt-subtot.tot-log = 0
           tt-subtot.tot-cnt = 0.
  end.
end procedure. /* DecodeOptionString */


/* Name : GetOutputFile
 * Desc : get the name of the output file
 */
procedure GetOutputFile:
  define output parameter cOutputFile as character no-undo.

  define variable fil-sp  as character  no-undo initial "".
  define variable fil-ext as character  no-undo.
  define variable lOk     as logical    no-undo.

  /* If output type is blank, allow choice via filename */
  if cFileType = tt-text  or cFileType = "" then fil-sp = fil-sp + ",*.txt":U.
  &if "{&opsys}":U = "WIN32":U &then
  if cFileType = tt-excel or cFileType = "" then fil-sp = fil-sp + ",*.xls":U.
  &endif
  if cFileType = tt-csv   or cFileType = "" then fil-sp = fil-sp + ",*.csv":U.
  if cFileType = tt-sylk  or cFileType = "" then fil-sp = fil-sp + ",*.slk":U.
  if cFileType = tt-df    or cFileType = "" then fil-sp = fil-sp + ",*.df":U.
  if cFileType = tt-xml   or cFileType = "" then fil-sp = fil-sp + ",*.xml":U.
  fil-sp = trim(fil-sp, {&sComma}).
  cOutputFile = pcFileName.

  /* Prompt for filename */
  if cOutputFile = "" and not cFileType = tt-browse then
  do:
    fil-ext = cFileType.
    if cFileType = "" then fil-ext = lc(tt-text).

    &if "{&opsys}":U = "WIN32":U &then
      /* Windows specific dialog, using filters */
      system-dialog get-file cOutputFile
        filters "Standard (":U + fil-sp + ")":U fil-sp,
                "Any File (*.*)":U              "*.*":U
        ask-overwrite
        create-test-file
        default-extension ".":U + fil-ext
        save-as
        update lOk.
    &else
      /* Generic dialog, without using filters. */
      run adeedit/_dlggetf.p
          ( input "Enter filename to save to",
            input true,
            input "",
            input-output cOutputFile,
            output lOk).
    &endif
    if not lOk then return.
  end.

  /* Process blank type choice */
  if cFileType = "" then
  do:
    fil-ext = entry(num-entries(cOutputFile, ".":U), cOutputFile, ".":U).
    case fil-ext:
      when tt-text  then cFileType = tt-text.
      when tt-excel then cFileType = tt-excel.
      when tt-csv   then cFileType = tt-csv.
      when tt-sylk  then cFileType = tt-sylk.
      when tt-df    then cFileType = tt-df.
      when tt-xml   then cFileType = tt-xml.
      otherwise          cFileType = tt-text.
    end case.
  end.
end procedure. /* GetOutputFile */


/* Name : SylkFormat
 * Desc : Add a part to the Sylk structure.
 */
procedure SylkFormat :
  define input parameter p-typ as character no-undo.
  define input parameter p-val as character no-undo.
  define input parameter p-fmt as character no-undo.
  define variable nxt as integer no-undo.

  find last tt-sylk where tt-sylk.typ = p-typ no-lock no-error.
  if available(tt-sylk) then
    nxt = tt-sylk.seq + 1.
  else
    nxt = 0.
  create tt-sylk.
  assign tt-sylk.typ = p-typ
         tt-sylk.seq = nxt
         tt-sylk.val = p-val
         tt-sylk.fmt = p-fmt.
end procedure. /* SylkFormat */


/* Name : tt-init-label
 * Desc : Initialize the query.
 */
procedure tt-init-label :
  iNumField = btt:num-fields.
  iNumIndex = 0.
  iCurField = 0.
  iSeqField = 0.
end procedure. /* tt-init-label */


/* Name : tt-next-label
 * Desc : Get the next query record.
 */
procedure tt-next-label :
  if iNumIndex > 0 then
  do: /* Bump index */
    iCurIndex = iCurIndex + 1.
    if iCurIndex > iNumIndex then iNumIndex = 0.
  end.
  if iNumIndex = 0 then
  do: /* Bump field */
    iCurField = iCurField + 1.
    if iCurField > iNumField then
    do:
      lEol = true.
      return.
    end.
    else
    do:
      hwf = btt:buffer-field(iCurField).
      iNumIndex = hwf:extent.
      if iNumIndex = 0 then
        iCurIndex = 0.
      else
        iCurIndex = 1.
    end.
  end.

  if iCurIndex = 0 then cFieldLabel = hwf:label.
                   else cFieldLabel = hwf:label + "[":U + string(iCurIndex) + "]":U.
  cFieldName   = getFieldName(hwf, iCurIndex).
  cFieldFormat = fill("X":U, getLabelWidth(hwf)).
  iSeqField    = iSeqField + 1.

  for each tt-subtot where tt-subtot.st-name = cFieldName exclusive-lock:
    case hwf:data-type:
      when "integer":U   then tt-subtot.tot-typ = {&tot-as-int}.
      when "decimal":U   then tt-subtot.tot-typ = {&tot-as-dec}.
      when "logical":U   then tt-subtot.tot-typ = {&tot-as-log}.
      when "character":U then tt-subtot.tot-typ = {&tot-as-cnt}.
      when "date":U      then tt-subtot.tot-typ = {&tot-as-cnt}.
    end case.
    assign tt-subtot.st-pos = 0
           tt-subtot.st-fmt = cFieldFormat.
  end.
  create tt-labels.
  assign tt-labels.bk-name = cFieldName
         tt-labels.st-lab  = cFieldLabel.
end procedure. /* tt-next-label */


/* Name : tt-init-query
 * Desc : Initialize the query.
 */
procedure tt-init-query :
  define variable lResult as logical   no-undo.

  if not valid-handle( hQuery ) then
  do:
    create query hQuery.
    hQuery:set-buffers(btt:handle).
  end.

  lResult = hQuery:query-prepare(substitute("FOR EACH &1 &2 &3":U
                                           , phSource:name
                                           , cWhere
                                           , cBreakBy )) no-error.
  if lResult = false then do:
    run fatalError (substitute("Invalid querystring passed to tt-file.p: &1 FOR EACH &2 &3 &4"
                              , {&lf}
                              , phSource:name
                              , cWhere
                              , cBreakBy )) .
    return error.

  end.

  hQuery:query-open.
  lEof = false.
  lEol = true.
  iNumField = btt:num-fields.
  iSeqField = 0.
end procedure. /* tt-init-query */


/* Name : tt-next-record
 * Desc : Get the next query record & check for breaks.
 */
procedure tt-next-record :
  define variable b-hwf   as handle    no-undo.
  define variable hTmpFld as handle    no-undo.
  define variable ndx     as integer   no-undo.
  define variable str     as character no-undo.
  define variable cDt     as character no-undo.
  define variable iExtBeg as integer   no-undo.
  define variable iExtEnd as integer   no-undo.
  define variable iExtCnt as integer   no-undo.

  if iRecordCounter > 0 then
  do:
    /* Accumulate by field */
    do ndx = 1 to iNumField:
      assign hwf = btt:buffer-field(ndx)
             cDt = hwf:data-type.

      if hwf:extent = 0 then assign iExtBeg = 0 iExtEnd = 0.
                        else assign iExtBeg = 1 iExtEnd = hwf:extent.

      do iExtCnt = iExtBeg to iExtEnd:
        str = getFieldName(hwf,iExtCnt).
        for each tt-subtot where tt-subtot.st-name = str exclusive-lock:
          if cDt = "integer":U and hwf:buffer-value(iExtCnt) <> ?    then tt-subtot.tot-int = tt-subtot.tot-int + hwf:buffer-value(iExtCnt).
          if cDt = "decimal":U and hwf:buffer-value(iExtCnt) <> ?    then tt-subtot.tot-dec = tt-subtot.tot-dec + hwf:buffer-value(iExtCnt).
          if cDt = "logical":U and hwf:buffer-value(iExtCnt)  = true then tt-subtot.tot-log = tt-subtot.tot-log + 1.
          if cDt = "character":U                                     then tt-subtot.tot-cnt = tt-subtot.tot-cnt + 1.
          if cDt = "date":U    and hwf:buffer-value(iExtCnt) <> ?    then tt-subtot.tot-cnt = tt-subtot.tot-cnt + 1.
        end.
      end.
    end.
    hTmpBuf:buffer-copy(btt).
  end.

  assign iNumIndex = 0
         iCurField = 0
         iSeqField = 0
         lEol      = false.

  hQuery:get-next().
  if hQuery:query-off-end then lEof = true.

  /* Generate change detection */
  cBufComp = "".
  if iRecordCounter > 0 then /* No change on first record */
  do:
    do ndx = 1 to iNumField:
      b-hwf = btt:buffer-field(ndx).
      hTmpFld = hTmpBuf:buffer-field(ndx).

      if b-hwf:extent = 0 then assign iExtBeg = 0 iExtEnd = 0.
                          else assign iExtBeg = 1 iExtEnd = b-hwf:extent.

      do iExtCnt = iExtBeg to iExtEnd:
        str = getFieldName(b-hwf, iExtCnt).
        find first tt-labels where tt-labels.bk-name = str exclusive-lock no-error.
        if available(tt-labels) then
          tt-labels.st-brk = trim(hTmpFld:buffer-value(iExtCnt)).
        find first tt-subtot where tt-subtot.bk-name = str no-lock no-error.
        if not available(tt-subtot) then next.
        if lEof or b-hwf:buffer-value(iExtCnt) <> hTmpFld:buffer-value(iExtCnt) then
          cBufComp = cBufComp + str + {&sComma}.
      end.
    end.
  end.
  cBufComp = trim(cBufComp, {&sComma}).

end procedure. /* tt-next-record */


/* Name : tt-next-field
 * Desc : Get the next field to display.
 */
procedure tt-next-field :
  if iNumIndex > 0 then
  do: /* Bump index */
    iCurIndex = iCurIndex + 1.
    if iCurIndex > iNumIndex then iNumIndex = 0.
  end.

  if iNumIndex = 0 then
  do: /* Bump field */
    iCurField = iCurField + 1.
    if iCurField > iNumField then
    do:
      lEol = true.
      return.
    end.
    else
    do:
      hwf = btt:buffer-field(iCurField).
      iNumIndex = hwf:extent.
      if iNumIndex = 0 then
        iCurIndex = 0.
      else
        iCurIndex = 1.
    end.
  end.

  if iCurIndex = 0 then cFieldName   = hwf:label.
                   else cFieldName   = hwf:label + "[":U + string(iCurIndex) + "]":U.
  cFieldFormat = fill("X":U, getLabelWidth(hwf)).
  iSeqField    = iSeqField + 1.
  cFieldStr    = hwf:string-value(iCurIndex).
  cFieldBuf    = hwf:buffer-value(iCurIndex).
end procedure. /* tt-next-field */


/* Name : tt-end-query
 * Desc : Write in DF style.
 */
procedure tt-end-query :
  hQuery:query-close().
  delete object hQuery.
end procedure. /* tt-end-query */


/* Name : df_format
 * Desc : Write in DF style.
 */
procedure df_format:
    output to value(cOutputFile).

    /* table */
    put unformatted 'ADD TABLE "':U   cTableName '"':U skip
                    '  AREA "Schema Area"':U skip
                    '  DUMP-NAME "':U substring(cTableName, 1, 8, {&sChar}) '"':U skip(2).
    /* field */
    do wf = 1 to nf:
      assign hwf = btt:buffer-field(wf).
      if FieldRequired(hwf) then
      do:
        put unformatted 'ADD FIELD "':U hwf:name '" OF "':U cTableName '" AS ':U hwf:data-type skip
                        '  FORMAT "':U hwf:format '"':U         skip
                        '  INITIAL "':U trim(hwf:initial) '"':U skip
                        '  POSITION ':U trim(string(wf + 1))    skip.
        if hwf:data-type = "character":U then
          put unformatted '  SQL-WIDTH ':U hwf:width-chars * 2 skip.
        else
          put unformatted '  SQL-WIDTH 4':U skip.
        put unformatted '  LABEL "':U hwf:label '"':U               skip
                        '  COLUMN-LABEL "':U hwf:column-label '"':U skip.
        if hwf:help <> ? then put unformatted '  HELP "':U hwf:help '"':U skip.
        put unformatted '  ORDER ':U + trim(string(wf * 10)) skip(2).
      end.
    end.

    /* index */
    do wf = 1 to 200:
      if btt:index-information(wf) = ? then leave.
      assign str = btt:index-information(wf).
      if entry(1, str) = "default":U then leave.
      put unformatted 'ADD INDEX "':U entry(1, str) '" ON "':U cTableName '"':U skip
                      '  AREA "Schema Area"':U skip.
      if entry(2, str) = "1":U then put unformatted '  UNIQUE':U  skip.
      if entry(3, str) = "1":U then put unformatted '  PRIMARY':U skip.
      if entry(4, str) = "1":U then put unformatted '  WORD':U    skip.
      do nx = 5 to num-entries(str) by 2:
        tag = "0":U.
        tag = entry(nx + 1, str) no-error.
        put unformatted '  INDEX-FIELD "':U entry(nx, str) '" ':U
          (if tag = "0":U then 'ASCENDING ':U else 'DESCENDING ':U) skip.
      end.
      put unformatted skip(1).
    end.

    /* Got this from: http://www.v9stuff.com/dynexport.htm, */
    /* thanks to Tony Lavinio and Peter van Dam             */
    assign nx = seek(output).
    put unformatted ".":U skip
                    "PSC":U skip
                    "cpstream=":U session:cpstream skip
                    ".":U skip
                    string(nx, "9999999999":U) skip.
    output close.
end procedure. /* df_format */


/* Name : txt_format
 * Desc : Write in TXT style.
 */
procedure txt_format:
    define variable cColumnSep           as character no-undo.
    define variable cCrossChar           as character no-undo.
    define variable cHorGridLine         as character no-undo.
    define variable lHorGridFieldChanged as logical   no-undo.
    define variable cHorGridFieldValue   as character no-undo.
    define variable cPrintLine           as character no-undo.
    define variable cTotalLine           as character no-undo.
    define variable cTotalValue          as character no-undo.
    define variable cBreakKey            as character no-undo.
    define variable cSpaceCharHor        as character no-undo.
    define variable cFillFormat          as character no-undo.
    define variable cnt                  as integer   no-undo.

    /* Set the column seperators to double wide grid horizontal size.
     * Add the vertical seperator if needed.
     */
    cSpaceCharHor = fill(" ":U, length(cGridCharHor, {&sChar})).
    if lVerGrid then
      assign
        cColumnSep = cSpaceCharHor + cGridCharVer   + cSpaceCharHor
        cCrossChar = cGridCharHor  + cGridCharCross + cGridCharHor.
    else
      assign
        cColumnSep = cSpaceCharHor + cSpaceCharHor
        cCrossChar = cGridCharHor  + cGridCharHor.

    if lAppend then output to value(cOutputFile) append.
               else output to value(cOutputFile).

    /* Make a line with the names of the columns */
    assign lFirstField  = true
           cHorGridLine = ""
           cPrevRec     = "".
    run tt-init-label.
    repeat: /* process field names */
      run tt-next-label.
      if lEol then leave.
      cPrevRec = cPrevRec + {&mk}.
      if FieldRequired(hwf) then
      do:
        if not lFirstField then
        do:
          cHorGridLine = cHorGridLine + cCrossChar.
          if lPrintLabels then put unformatted cColumnSep.
        end.
        lFirstField = false.
        if lPrintLabels then
          put unformatted string(display_label(cFieldLabel), cFieldFormat).
        cHorGridLine = cHorGridLine + string(fill(cGridCharHor, getLabelWidth(hwf)), cFieldFormat).
      end. /* field required in output */
    end. /* process field names */

    if lPrintLabels then
      put unformatted {&lf}.

    run tt-init-query.

    assign
      cHorGridFieldValue = ""
      iRecordCounter     = 0.

    repeat: /* process table */
      run tt-next-record.
      /* Check for a subtotal break */
      if (cBufComp <> "" and cSubBreak <> "") or (lEof and cSubTotal <> "") then
      do:
        /* Find break level */
        brk-new = num-entries(cSubBreak).
        if lEof then
          brk-new = 0.
        else
        do cnt = 1 to num-entries(cBufComp):
          find first tt-subtot where tt-subtot.bk-name = entry(cnt, cBufComp)
                                     no-lock no-error.
          if available(tt-subtot) then
            if tt-subtot.lvl < brk-new then brk-new = tt-subtot.lvl.
        end.
        if brk-new > brk-old then
          brk-old = brk-new.
        do brk-cur = brk-old to brk-new by -1:
          /* Print subtotals */
          cTotalLine = "".
          cBreakKey = "".
          find first tt-labels where tt-labels.bk-name = entry(brk-cur, cSubBreak)
                                     no-lock no-error.
          if available(tt-labels) then cBreakKey = tt-labels.st-lab + " " + tt-labels.st-brk.
          for each tt-subtot where tt-subtot.lvl = brk-cur no-lock by tt-subtot.st-pos:
            cnt = length(cTotalLine, {&sChar}).
            if tt-subtot.st-pos > cnt then
              cTotalLine = cTotalLine + fill(" ":U, tt-subtot.st-pos - cnt).
            case tt-subtot.tot-typ:
              when {&tot-as-int} then cTotalValue = string(string(tt-subtot.tot-int), tt-subtot.st-fmt).
              when {&tot-as-dec} then cTotalValue = string(string(tt-subtot.tot-dec), tt-subtot.st-fmt).
              when {&tot-as-log} then cTotalValue = string(string(tt-subtot.tot-log), tt-subtot.st-fmt).
              when {&tot-as-cnt} then cTotalValue = string(string(tt-subtot.tot-cnt), tt-subtot.st-fmt).
            end case.
            cTotalLine = cTotalLine + cTotalValue.
          end.
          put unformatted cTotalLine.
          if brk-cur > 0 then put unformatted cColumnSep cBreakKey.
                         else put unformatted cColumnSep "Grand Total":U.
          put unformatted {&lf}.
        end.
        /* Reset subtotals */
        for each tt-subtot where tt-subtot.lvl >= brk-new exclusive-lock:
          assign tt-subtot.tot-int = 0
                 tt-subtot.tot-dec = 0.0
                 tt-subtot.tot-log = 0
                 tt-subtot.tot-cnt = 0.
        end.
        if brk-new > 0 or (brk-old > 0 and cSubBreak <> "") then
          put unformatted {&lf}.
      end.
      if lEof then leave.
      assign
        lFirstField          = true
        cThisRec             = ""
        cPrintLine           = ""
        lHorGridFieldChanged = no
        iPrnField            = 0
        .
      repeat: /* process record */
        run tt-next-field.
        if lEol then leave.
        if FieldRequired(hwf) then
        do:
          iPrnField = iPrnField + 1.
          if not lFirstField then
            cPrintLine = cPrintLine + cColumnSep.
          str = cFieldStr.
          if cFieldBuf = ? then str = "?":U.
          /* Record value for comparisons */
          cThisRec = cThisRec + str + {&mk}.

          /* ptpt */
          /* See if the field which determines the horizontal grid line, has changed its value */
          if     cHorGridFieldName   = cFieldName
             and cHorGridFieldValue <> str then
            assign lHorGridFieldChanged = true
                   cHorGridFieldValue   = str.

          /* Save print position for subtotals */
          if iRecordCounter = 0 and cSubTotal <> "" then
          do:
            for each tt-subtot where tt-subtot.st-name = cFieldName exclusive-lock:
              assign tt-subtot.st-pos = length(cPrintLine, {&sChar}).
            end.
          end.
          /* check on differences with prev record */
          if iRecordCounter > 0
            and can-do(cFirstOfList, cFieldName)
            and str = entry(iSeqField, cPrevRec, {&mk}) then
            str = "".
          cPrintLine = cPrintLine + string(str, cFieldFormat).
          lFirstField = false.
        end. /* field required in output */
        else
          cThisRec = cThisRec + {&mk}.
      end. /* process record */

      /* horizontal line? */
      if not (iRecordCounter = 0 and lPrintLabels = false) then
        if (    iGridInterval = 0
            and lHorGridFieldChanged = true)
        or (    iGridInterval > 0
            and lHorGrid = true
            and iRecordCounter MODULO iGridInterval = 0) then
        put unformatted cHorGridLine {&lf}.

      /* print the data */
      put unformatted
        right-trim(cPrintLine) {&lf}.

      /* remember 'prev' record and increase counter */
      assign
        cPrevRec       = cThisRec
        iRecordCounter = iRecordCounter + 1.
    end. /* process table */

    /* horizontal line? */
    if   (iGridInterval = 0)
      or (iGridInterval > 0 and lHorGrid = true and iRecordCounter MODULO iGridInterval = 0) then
      put unformatted cHorGridLine {&lf}.

    run tt-end-query.
    output close.
end procedure. /* txt_format */


&if "{&opsys}":U = "WIN32":U &then
/* Name : xls_format
 * Desc : Write in XLS style.
 */
procedure xls_format:
    define variable chExcelApplication   as com-handle no-undo.
    define variable chWorkbook           as com-handle no-undo.
    define variable chWorksheet          as com-handle no-undo.
    define variable chRange              as com-handle no-undo.
    define variable cols                 as character  no-undo.
    define variable cnt                  as integer    no-undo.
    define variable cTotalValue          as character  no-undo.
    define variable NS                   as integer    no-undo.
    define variable chTest               as com-handle no-undo.
    define variable cSheetName           as character  no-undo.
    define variable cBreakKey            as character  no-undo.

    create "Excel.Application":U chExcelApplication.
    assign chExcelApplication:Visible = lExcelVisible.
    if lAppend then
    do:
      chExcelApplication:Workbooks:Open(cOutputFile).
      NS = chExcelApplication:Sheets:Count.
      assign chRange     = chExcelApplication:Sheets:item(NS).
      assign chWorkSheet = chExcelApplication:Sheets:Add(, chRange,, {&xlWBATWorksheet}).
      assign chWorkbook  = chExcelApplication:WorkBooks:Item(1).
      release object chRange no-error.
    end.
    else
    do:
      chExcelApplication:Workbooks:Add({&xlWBATWorksheet}).
      NS = chExcelApplication:Sheets:Count.
      assign chWorkSheet = chExcelApplication:Sheets:Item(1).
      assign chWorkbook  = chExcelApplication:WorkBooks:Item(1).
    end.
    NS = NS + 1.
    if cBrowseTitle <> "" then cSheetName = getValidFileName(cBrowseTitle).
                          else cSheetName = getValidFileName(cOutputFile).
    assign chTest = chExcelApplication:Sheets:item(cSheetName) no-error.
    if valid-handle(chTest) then
      cSheetName = getValidFileName(string(NS) + " ":U + cSheetName).
    release object chTest no-error.
    assign chWorkSheet:Name = cSheetName.
    assign
      nc          = 0
      cThisRec    = "".

    run tt-init-label.
    repeat: /* process field names */
      run tt-next-label.
      if lEol then leave.
      if FieldRequired(hwf) then
      do:
        nc = nc + 1.
        cols = getExcelColumn(nc).
        if lPrintLabels then do:
          chWorksheet:Range(cols + {&sColon} + cols):ColumnWidth = getExcelColumnWidth(hwf).
          chWorksheet:Range(cols + "1":U):Value = display_label(cFieldLabel).
        end.
        case hwf:data-type:
          when "integer":U or
          when "date":U    or
          when "decimal":U then
            chWorksheet:Range(cols + "1":U):HorizontalAlignment = {&xlRight}.
        end case.
      end. /* field required in output */
    end. /* for each field */

    run tt-init-query.

    assign
      iRecordCounter = 0.

    repeat: /* process table */
      run tt-next-record.
      /* Check for a subtotal break */
      if (cBufComp <> "" and cSubBreak <> "") or (lEof and cSubTotal <> "") then
      do:
        /* Find break level */
        brk-new = num-entries(cSubBreak).
        if lEof then
          brk-new = 0.
        else
        do cnt = 1 to num-entries(cBufComp):
          find first tt-subtot where tt-subtot.bk-name = entry(cnt, cBufComp)
                                     no-lock no-error.
          if available(tt-subtot) then
            if tt-subtot.lvl < brk-new then brk-new = tt-subtot.lvl.
        end.
        if brk-new > brk-old then
          brk-old = brk-new.
        do brk-cur = brk-old to brk-new by -1:
          cBreakKey  = "".
          find first tt-labels where tt-labels.bk-name = entry(brk-cur, cSubBreak)
                                     no-lock no-error.
          if available(tt-labels) then cBreakKey = tt-labels.st-lab + " " + tt-labels.st-brk.
          /* Print subtotals */
          for each tt-subtot where tt-subtot.lvl = brk-cur no-lock by tt-subtot.st-pos:
            case tt-subtot.tot-typ:
              when {&tot-as-int} then cTotalValue = string(tt-subtot.tot-int).
              when {&tot-as-dec} then cTotalValue = string(tt-subtot.tot-dec).
              when {&tot-as-log} then cTotalValue = string(tt-subtot.tot-log).
              when {&tot-as-cnt} then cTotalValue = string(tt-subtot.tot-cnt).
              otherwise               cTotalValue = "".
            end case.
            cols = getExcelColumn(tt-subtot.st-pos) + string(iRecordCounter + 2).
            chWorksheet:Range(cols):Value = right-trim(cTotalValue).
          end.
          if brk-cur > 0 then cTotalValue = cBreakKey.
                         else cTotalValue = "Grand Total":U.
          if cSubTotal <> "" or (cSubBreak <> "" and brk-cur > 0) then
          do:
            cols = getExcelColumn(nc + 1) + string(iRecordCounter + 2).
            chWorksheet:Range(cols):Value = right-trim(cTotalValue).
            iRecordCounter = iRecordCounter + 1.
          end.
        end.
        /* Reset subtotals */
        for each tt-subtot where tt-subtot.lvl >= brk-new exclusive-lock:
          assign tt-subtot.tot-int = 0
                 tt-subtot.tot-dec = 0.0
                 tt-subtot.tot-log = 0
                 tt-subtot.tot-cnt = 0.
        end.
        if brk-new > 0 or (brk-old > 0 and cSubBreak <> "") then
          iRecordCounter = iRecordCounter + 1.
      end.
      if lEof then leave.
      assign
        nc          = 0
        cThisRec    = "".

      repeat: /* process fields */
        run tt-next-field.
        if lEol then leave.
        if FieldRequired(hwf) then
        do:
          nc = nc + 1.
          str = cFieldStr.
          if cFieldBuf = ? then str = "?":U.
          /* Record value for comparisons */
          cThisRec = cThisRec + str + {&mk}.
          /* Save print position for subtotals */
          if iRecordCounter = 0 and cSubTotal <> "" then
          do:
            for each tt-subtot where tt-subtot.st-name = cFieldName exclusive-lock:
              assign tt-subtot.st-pos = nc.
            end.
          end.
          if iRecordCounter = 0 and cSubBreak <> "" then
          do:
            for each tt-subtot where tt-subtot.bk-name = cFieldName exclusive-lock:
              assign tt-subtot.st-pos = nc.
            end.
          end.
          /* check on differences with prev record (not for array fields) */
          if iRecordCounter > 0
            and can-do(cFirstOfList, cFieldName)
            and str = entry(iSeqField, cPrevRec, {&mk}) then
            str = "".
          cols = getExcelColumn(nc) + string(iRecordCounter + 2).
          chWorksheet:Range(cols):Value = right-trim(str).
        end. /* Field required in output */
        else
          cThisRec = cThisRec + {&mk}.
      end. /* for each field */

      /* remember 'prev' record */
      cPrevRec = cThisRec.
      iRecordCounter = iRecordCounter + 1.

    end. /* repeat */
    run tt-end-query.

    /* Adjust column sizes */
    cols = getExcelColumn(nc).
    chRange = chWorksheet:Range("A:":U + cols):Columns().
    chRange:AutoFit().
    release object chRange no-error.

    if lPrintLabels then
    do:
      /* Set first row as title row */
      chWorksheet:Range("A1:A1":U):Select.
      chWorkbook:Windows(1):SplitColumn = 0.
      chWorkbook:Windows(1):SplitRow    = 1.
      chWorkbook:Windows(1):FreezePanes = True.
    end.

    /* Do we want interaction with excel? */
    chExcelapplication:DisplayAlerts = lExcelAlert.

    /* Save data */
    if chExcelapplication:Version begins "8":U
      then chWorkBook:SaveAs(cOutputFile, {&xlExcel9795},,,,,,).
      else chWorkBook:SaveAs(cOutputFile, {&xlWorkbookNormal},,,,,,).

    chWorkBook:Close().
    release object chWorkSheet  no-error.
    release object chWorkBook   no-error.
    chExcelApplication:Quit().
    release object chExcelApplication no-error.
end procedure. /* xls_format */
&endif


/* Name : csv_format
 * Desc : Write in CSV style.
 */
procedure csv_format:
  define variable cnt         as integer   no-undo.
  define variable cTotalLine  as character no-undo.
  define variable cBreakKey   as character no-undo.
  define variable cTotalValue as character no-undo.
  define variable cPrintLine  as character no-undo.

    if lAppend then output to value(cOutputFile) append.
               else output to value(cOutputFile).

    /* Make a line with the names of the columns */
    lFirstField = true.
    cPrevRec    = "".
    run tt-init-label.
    repeat: /* process field names */
      run tt-next-label.
      if lEol then leave.
      cPrevRec = cPrevRec + {&mk}.
      if FieldRequired(hwf) then
      do:
        if lPrintLabels and not lFirstField then put unformatted {&sComma}.
        lFirstField = false.
        if lPrintLabels then
          put unformatted {&qt} trim(display_label(cFieldLabel)) {&qt}.
      end. /* Field required in output */
    end. /* process field names */

    if lPrintLabels then
      put unformatted {&lf}.

    run tt-init-query.

    assign
      iRecordCounter = 0.

    repeat: /* process table */
      run tt-next-record.
      /* Check for a subtotal break */
      if (cBufComp <> "" and cSubBreak <> "") or (lEof and cSubTotal <> "") then
      do:
        /* Find break level */
        brk-new = num-entries(cSubBreak).
        if lEof then
          brk-new = 0.
        else
        do cnt = 1 to num-entries(cBufComp):
          find first tt-subtot where tt-subtot.bk-name = entry(cnt, cBufComp)
                                     no-lock no-error.
          if available(tt-subtot) then
            if tt-subtot.lvl < brk-new then brk-new = tt-subtot.lvl.
        end.
        if brk-new > brk-old then
          brk-old = brk-new.
        do brk-cur = brk-old to brk-new by -1:
          /* Print subtotals */
          cTotalLine = "".
          cBreakKey  = "".
          iPrnField  = 1.
          find first tt-labels where tt-labels.bk-name = entry(brk-cur, cSubBreak)
                                     no-lock no-error.
          if available(tt-labels) then cBreakKey = tt-labels.st-lab + " " + tt-labels.st-brk.
          for each tt-subtot where tt-subtot.lvl = brk-cur no-lock by tt-subtot.st-pos:
            if tt-subtot.st-pos > iPrnField then
              cTotalLine = cTotalLine + fill({&sComma}, tt-subtot.st-pos - iPrnField).
            iPrnField = tt-subtot.st-pos.
            case tt-subtot.tot-typ:
              when {&tot-as-int} then cTotalValue = string(tt-subtot.tot-int).
              when {&tot-as-dec} then cTotalValue = string(tt-subtot.tot-dec).
              when {&tot-as-log} then cTotalValue = string(tt-subtot.tot-log).
              when {&tot-as-cnt} then cTotalValue = string(tt-subtot.tot-cnt).
            end case.
            cTotalLine = cTotalLine + cTotalValue.
          end.
          cTotalLine = cTotalLine + fill({&sComma}, num-entries(cPrevRec, {&mk}) - iPrnField).
          put unformatted cTotalLine.
          if brk-cur > 0 then cTotalValue = cBreakKey.
                         else cTotalValue = "Grand Total":U.
          put unformatted cTotalValue {&lf}.
          if brk-new > 0 or (brk-old > 0 and cSubBreak <> "") then
            put unformatted '""':U {&lf}.
        end.
        /* Reset subtotals */
        for each tt-subtot where tt-subtot.lvl >= brk-new exclusive-lock:
          assign tt-subtot.tot-int = 0
                 tt-subtot.tot-dec = 0.0
                 tt-subtot.tot-log = 0
                 tt-subtot.tot-cnt = 0.
        end.
      end.
      if lEof then leave.
      assign
        lFirstField = true
        cThisRec    = ""
        iPrnField   = 0
        cPrintLine  = ""
        .
      repeat: /* process record */
        run tt-next-field.
        if lEol then leave.
        if FieldRequired(hwf) then
        do:
          iPrnField = iPrnField + 1.
          if not lFirstField then cPrintLine = cPrintLine + {&sComma}.
          str = cFieldStr.
          if cFieldBuf = ? then str = "?":U.
          /* Record value for comparisons */
          cThisRec = cThisRec + str + {&mk}.
          /* Save print position for subtotals */
          if iRecordCounter = 0 and cSubTotal <> "" then
          do:
            for each tt-subtot where tt-subtot.st-name = cFieldName exclusive-lock:
              assign tt-subtot.st-pos = iPrnField.
            end.
          end.
          /* check on differences with prev record (not for array fields) */
          if iRecordCounter > 0
            and can-do(cFirstOfList, cFieldName)
            and str = entry(iSeqField, cPrevRec, {&mk}) then
            str = "".
          /* Quote character fields */
          if hwf:data-type = "character":U then
            str = {&qt} + right-trim(str) + {&qt}.
          cPrintLine = cPrintLine + trim(str).
          lFirstField = false.
        end. /* Field required in output */
        else
          cThisRec = cThisRec + {&mk}.
      end. /* process record */

      /* print the data */
      put unformatted
        cPrintLine {&lf}.

      /* remember 'prev' record */
      cPrevRec = cThisRec.
      iRecordCounter = iRecordCounter + 1.

    end. /* process table */
    run tt-end-query.
    output close.
end procedure. /* csv_format */


/* Name : getTooltip
 * Desc : Return a browse row as a string.
 */
function getTooltip returns character ( hPar as handle, iRowNr as integer):
  define variable hLocalBuffer as handle    no-undo.
  define variable hLocalField  as handle    no-undo.
  define variable cString      as character no-undo.
  define variable iFieldNr     as integer   no-undo.
  define variable iExtentNr    as integer   no-undo.

  hPar:fetch-selected-row(iRowNr) no-error.
  if not hPar:query:get-current(no-lock) then return no-apply.
  hLocalBuffer = hPar:query:get-buffer-handle(1).

  do iFieldNr = 1 to hLocalBuffer:num-fields:
    hLocalField = hLocalBuffer:buffer-field(iFieldNr).

    if not FieldRequired(hLocalField) then next.
    if iFieldNr > 1 then cString = cString + {&lf}.
    if hLocalField:extent = 0 then
    do:
      cString = cString + display_label(hLocalField:label) + ": ":U
              + ( if hLocalField:buffer-value = ? then '?':U else hLocalField:string-value).
    end.
    else
    do iExtentNr = 1 to hLocalField:extent:
      if iExtentNr = 1 then cString = cString + display_label(hLocalField:label) + ": ":U.
                       else cString = cString + ", ":U.
      cString = cString
              + ( if hLocalField:buffer-value(iExtentNr) = ? then '?':U else hLocalField:string-value(iExtentNr)).
    end. /* iExtent */
  end. /* iField */

  return cString.
end function. /* getTooltip */


/* Name : brs_format
 * Desc : Create a browse window.
 */
procedure brs_format:
  define variable bwin             as handle    no-undo.
  define variable hq               as handle    no-undo.
  define variable hb               as handle    no-undo.
  define variable iCalculatedWidth as integer   no-undo.

  define sub-menu m_File
    menu-item m_Exit   label "E&xit":U
    menu-item m_Clip   label "&Clip":U
    menu-item m_Choose label "C&hoose":U.

  define menu   m_Bar  menubar sub-menu m_File label "&File":U.
  define button btExit   label "E&xit":U   size 12 by .92.
  define button btClip   label "&Clip":U   size 12 by .92.
  define button btChoose label "C&hoose":U size 12 by .92.

  define variable hBrowse as handle    no-undo.

  define frame fMaster
    btExit   at row 1 col 2
    btClip   at row 1 col 15
    btChoose at row 1 col 28
    with 1 down no-box keep-tab-order overlay
         side-labels no-underline three-d
         at col 1 row 1 size 90 by 15.

  run tt-init-query.

  create browse hBrowse
    assign
      query                  = hQuery
      frame                  = frame fMaster:handle
      row                    = 2
      columns                = 2
      width-chars            = 88
      height-chars           = 13
      row-markers            = false
      separators             = true
      expandable             = true
      no-validate            = true
      multiple               = lMultiple
      allow-column-searching = true
      column-scrolling       = false
      .

  iCalculatedWidth = 7.

  do wf = 1 to btt:num-fields:
    hwf = btt:buffer-field(wf).

    if FieldRequired(hwf) then
    do:
      nx  = hwf:extent.
      if nx = 0 then nx = 1.
      iCalculatedWidth  = iCalculatedWidth + (getColumnWidth(hwf) * nx).
      hBrowse:width-chars = minimum(320,iCalculatedWidth).
      hBrowse:add-like-column(hwf).
    end. /* Field required in output */
  end. /* for each field */

  /* If a value has been specified, use it. Otherwise use the calculated value */
  if iBrowseWidth = ?                     then iBrowseWidth = {&BrowseDefaultWidth}. /* default width */
  if iBrowseWidth > iCalculatedWidth      then iBrowseWidth = iCalculatedWidth.      /* unless smaller than default */
  if iBrowseWidth < {&BrowseMinimumWidth} then iBrowseWidth = {&BrowseMinimumWidth}. /* correct for max width */
  if iBrowseWidth > {&BrowseMaximumWidth} then iBrowseWidth = {&BrowseMaximumWidth}. /* correct for min width (room for buttons) */

  frame fMaster:width-chars          = iBrowseWidth + 1.
  frame fMaster:virtual-width-chars  = iBrowseWidth + 1.
  frame fMaster:height-chars         = iBrowseHeight.
  frame fMaster:virtual-height-chars = iBrowseHeight.

  hBrowse:width-chars = iBrowseWidth.
  hBrowse:height      = iBrowseHeight - 1.5.
  hBrowse:sensitive   = true.

  hQuery:query-open.

  create window bwin assign
    hidden             = yes
    title              = cBrowseTitle
    height             = iBrowseHeight
    width              = iBrowseWidth + 2
    max-height         = iBrowseHeight
    max-width          = iBrowseWidth + 2
    virtual-height     = iBrowseHeight
    virtual-width      = iBrowseWidth + 2
    resize             = no
    scroll-bars        = no
    status-area        = no
    bgcolor            = ?
    fgcolor            = ?
    keep-frame-z-order = yes
    three-d            = yes
    message-area       = no
    sensitive          = yes.

  assign bwin:menubar = menu m_Bar:handle.
  bwin:hidden = no.
  bwin:top-only = yes.

  on start-search of hBrowse
  do:
    define variable rCurrentRecord as rowid no-undo.
    define variable hLocalBuffer as handle    no-undo.

    /* Get the name of the column that is clicked */
    cSortColumn = self:current-column:name.

    /* Column-sorting allowed for this column? */
    if not can-do(cColumnSortList, cSortColumn) then do:
      message 'Sorting is disabled for this column.'
        view-as alert-box information buttons ok.
      return no-apply.
    end.

    /* Set way of sorting */
    if cSortColumn <> cSortColumnPrev then cSortMethod = 'ascending'.
    else
      cSortMethod = (if cSortMethod = 'ascending' then 'descending' else 'ascending').

    /* Compose the break-by clause */
    cBreakBy = "by " + cSortColumn + (if cSortMethod = 'ascending' then '' else ' descending').

    /* remember selected record */
    hBrowse:fetch-selected-row(1) no-error.
    if not hBrowse:query:get-current(no-lock) then return no-apply.
    hLocalBuffer = hBrowse:query:get-buffer-handle(1).

    rCurrentRecord = hLocalBuffer:rowid.

    run tt-init-query.
    hBrowse:query:reposition-to-rowid( rCurrentRecord ).

    cSortColumnPrev = cSortColumn.
  end.

  on entry of bwin
  do:
    on esc bell.
    on return tab.
  end.

  on leave of bwin
  do:
    on esc end-error.
    on return return.
  end.

  on esc, endkey, end-error, window-close of bwin anywhere do:
    apply "close":U to bwin.
    return no-apply.
  end.

  on choose of btClip /* Clip */
  or choose of menu-item m_Clip
  do:
    define variable iRow as integer   no-undo.

    output to "clipboard":U.
    do iRow = 1 to hBrowse:num-selected-rows:
      put unformatted replace( getTooltip(hBrowse,iRow), {&lf}, {&cr} + {&lf}) {&cr} {&lf}.
    end.
    output close.
    apply "entry":U to hBrowse.
  end.

  on choose of btExit /* Exit */
  or choose of menu-item m_Exit
  do:
    apply "close":U to bwin.
  end.

  on choose of btChoose /* Choose */
  or choose of menu-item m_Choose
  or default-action of hBrowse
  do:
    define variable iExtBeg as integer   no-undo.
    define variable iExtEnd as integer   no-undo.
    define variable iExtCnt as integer   no-undo.
    define variable iRowNr  as integer   no-undo.

    assign cReturnString = "".

    /* Provide a default value for the fieldname to be returned (browse) */
    if cReturnList = '':u then
      cReturnList = btt:buffer-field(1):name.

    do iRowNr = 1 to hBrowse:num-selected-rows:
      hBrowse:fetch-selected-row(iRowNr) no-error.
      if not hBrowse:query:get-current(no-lock) then return no-apply.
      hb = hBrowse:query:get-buffer-handle(1).

      do wf = 1 to num-entries(cReturnList):
        hwf = btt:buffer-field(entry(wf,cReturnList)) no-error.
        if not valid-handle(hwf) then next.

        if hwf:extent = 0 then
        do:
          cReturnString = left-trim(substitute("&1&2&3"
                                              , cReturnString
                                              , {&mk}
                                              , (if hwf:buffer-value = ? then "?":U else trim(hwf:string-value))
                                              ), {&mk}).
        end.
        else
        do nx = 1 to hwf:extent:
          cReturnString = left-trim(substitute("&1&2&3"
                                              , cReturnString
                                              , {&mk}
                                              , (if hwf:buffer-value(nx) = ? then "?":U else trim(hwf:string-value(nx)))
                                              ), {&mk}).
        end. /* Extent */
      end. /* wf */
    end. /* iRowNr */

    apply "close":U to bwin.
  end.

  on value-changed of hBrowse
  do:
    define variable iRowNr       as integer   no-undo.
    define variable cTempTooltip as character no-undo.

    if hBrowse:num-selected-rows = 1 then
      cTempTooltip = getTooltip(hBrowse,1).
    else
      cTempTooltip = substitute("&1 rows selected", hBrowse:num-selected-rows).

    if lBrowseTooltip then hBrowse:tooltip = cTempTooltip.
  end.

  enable btClip btExit btChoose with frame fMaster in window bwin.

  view frame fMaster in window bwin.
  view bwin.
  hBrowse:select-row(1).
  apply "entry":U to hBrowse.
  apply "value-changed":U to hBrowse.

  /* Sorry Greg */
  wait-for close of bwin.

  /* Clean up */
  hQuery:query-close().
  delete object hQuery.
  hQuery = ?.
  delete object hBrowse.
  hBrowse = ?.

  on esc end-error.
  on return return.
  delete widget bwin.
end procedure. /* brs_format */


/* Name : slk_format
 * Desc : Write in SYLK style.
 */
procedure slk_format:
  define variable fmtstr      as character no-undo.
  define variable fmtkey      as character no-undo.
  define variable fmttmp      as character no-undo.
  define variable fmtcol      as integer   no-undo.
  define variable fmtrow      as integer   no-undo.
  define variable fmtsiz      as integer   no-undo.
  define variable fmtwid      as character no-undo.
  define variable cnt         as integer   no-undo.
  define variable cTotalLine  as character no-undo.
  define variable cBreakKey   as character no-undo.
  define variable cTotalValue as character no-undo.

  /* Prefill temp table with formats and fonts */
  run SylkFormat({&sylk-fil-picture}, "PGeneral":U,         "").
  run SylkFormat({&sylk-fil-picture}, "Pdd\-mmm\-yy":U,     {&sylk-xls-dmy}).
  run SylkFormat({&sylk-fil-picture}, "Pmm/dd/yy":U,        {&sylk-xls-mdy}).
  run SylkFormat({&sylk-fil-picture}, "Pmmmm\ d\,\ yyyy":U, {&sylk-xls-md-y}).
  run SylkFormat({&sylk-fil-picture}, "Pyymmdd":U,          {&sylk-xls-ymd}).
  run SylkFormat({&sylk-fil-picture}, "Pyyyymmdd":U,        {&sylk-xls-cymd}).
  run SylkFormat({&sylk-fil-picture}, "Ph:mm":U,            {&sylk-xls-hm}).
  run SylkFormat({&sylk-fil-picture}, "Ph:mm\ AM/PM":U,     {&sylk-xls-hm-a}).
  run SylkFormat({&sylk-fil-picture}, "Ph:mm:ss":U,         {&sylk-xls-hms}).
  run SylkFormat({&sylk-fil-picture}, "Ph:mm:ss\ AM/PM":U,  {&sylk-xls-hms-a}).
  run SylkFormat({&sylk-fil-font},    "ECourier;M200;SB":U, {&sylk-xls-title}).
  run SylkFormat({&sylk-fil-font},    "ECourier;M200":U,    {&sylk-xls-body}).

  fmtcol   = 0.
  fmtkey   = "".
  cPrevRec = "".
  run tt-init-label.
  repeat:
    run tt-next-label.
    if lEol then leave.
    cPrevRec = cPrevRec + {&mk}.
    if FieldRequired(hwf) then
    do:
      fmtstr = hwf:format.
      /* Determine type of data in column */
      case hwf:data-type:
        when "integer":U   then fmttmp = {&SylkFormat-number}. /* Number  */
        when "decimal":U   then fmttmp = {&SylkFormat-number}. /* Number  */
        when "character":U then fmttmp = {&SylkFormat-string}. /* String  */
        when "date":U      then fmttmp = {&SylkFormat-date}.   /* Date    */
        otherwise               fmttmp = " ":U.                /* Unknown */
      end case.

      /* Determine format to use... */
      /* ...for a number */
      if     (   hwf:data-type = "integer":U
              or hwf:data-type = "decimal":U)
         and not fmtstr begins {&sColon} then
      do:
        fmtstr = replace(fmtstr, "9":U, "0":U). /* Forced character   */
        fmtstr = replace(fmtstr, "Z":U, "0":U). /* Forced character   */
        fmtstr = replace(fmtstr, ">":U, "#":U). /* Optional character */
        fmtstr = "P":U + fmtstr.
        /* Add format if not already there */
        if SylkValue({&sylk-fil-picture}, fmtstr) = "P0":U then
          run SylkFormat({&sylk-fil-picture}, fmtstr, fmtstr).
      end.

      /* ...for a time */
      if hwf:data-type = "integer":U then
      do:
        /* check for time formats */
        case hwf:format:
          when {&tt-fmt-hm}    then do: fmtstr = {&sylk-xls-hm}.    fmttmp = {&SylkFormat-time}. end.
          when {&tt-fmt-hm-a}  then do: fmtstr = {&sylk-xls-hm-a}.  fmttmp = {&SylkFormat-time}. end.
          when {&tt-fmt-hms}   then do: fmtstr = {&sylk-xls-hms}.   fmttmp = {&SylkFormat-time}. end.
          when {&tt-fmt-hms-a} then do: fmtstr = {&sylk-xls-hms-a}. fmttmp = {&SylkFormat-time}. end.
        end case.
      end.

      /* ...for a date */
      if hwf:data-type = "date":U then
      do:
        if fmtstr = {&tt-fmt-mdy}  then fmtstr = {&sylk-xls-mdy}.
        if fmtstr = {&tt-fmt-dmy}  then fmtstr = {&sylk-xls-dmy}.
        if fmtstr = {&tt-fmt-md-y} then fmtstr = {&sylk-xls-md-y}.
        if fmtstr = {&tt-fmt-ymd}  then fmtstr = {&sylk-xls-ymd}.
        if fmtstr = {&tt-fmt-cymd} then fmtstr = {&sylk-xls-cymd}.
      end.

      /* ...for a logical */
      if hwf:data-type = "logical":U then
      do:
        fmtstr = fill("X":U,
                      max(length(entry(1, fmtstr, "/":U), {&sChar}),
                          length(entry(2, fmtstr, "/":U), {&sChar}))).
        fmttmp = {&SylkFormat-logical}.
      end.

      fmtkey = fmtkey + fmttmp.
      /* Get data for column formatting and titles */
      fmtcol = fmtcol + 1.
      fmtwid = string(fmtcol).
      /* Get column width as max(label width, data width) */
      fmtsiz = length(cFieldLabel, {&sChar}).
      if fmttmp = {&SylkFormat-logical} then fmtsiz = max(fmtsiz, length(fmtstr, {&sChar})).
      if fmttmp = {&SylkFormat-time}    then fmtsiz = max(fmtsiz, length(fmtstr, {&sChar})).
                                        else fmtsiz = max(fmtsiz, length(cFieldFormat, {&sChar})).
      if fmttmp = {&SylkFormat-date} and fmtstr = {&sylk-xls-md-y} then
        fmtsiz = max(fmtsiz, 18).
      /* Store width info */
      run SylkFormat({&sylk-fil-c-wid},
                   "W":U + fmtwid + " ":U + fmtwid + " ":U + string(fmtsiz),
                   fmtwid).
      /* Store title info (alignment and text) */
      run SylkFormat({&sylk-fil-c-title},
                   "F;FG0":U + (if fmttmp={&SylkFormat-string} then "L":U else "R":U) + ";SD":U
                 + SylkValue({&sylk-fil-font}, {&sylk-xls-title}) + ";Y1;X":U + fmtwid + {&lf}
                 + "C;K":U + {&qt} + sylk-str(trim(display_label(cFieldLabel))) + {&qt},
                   fmtwid).
      /* Store format info */
      run SylkFormat({&sylk-fil-c-fmt},
                   SylkValue({&sylk-fil-picture}, fmtstr) + ";FG0G;C":U + fmtwid,
                   fmtwid).
    end. /* Field required in output */
    else
      fmtkey = fmtkey + " ":U.
  end.

  /* Write out file */
  output to value(cOutputFile).
  /* BNF: "ID" ";P" <name> ";N" ";E"
   * ;P <name>   authoring-program
   * ;N          File uses ;N style cell protection (not ;P style).
   * ;E          NE records are redundant as formulas support external ref's directly.
   */
  put unformatted "ID;PTT-FILE;N;E":U {&lf}.
  /* BNF: "P;P" <xl-pic>
   * "P;P" <xl-pic>   Excel style picture format.  Count from zero.
   */
  for each tt-sylk where tt-sylk.typ = {&sylk-fil-picture} no-lock by tt-sylk.seq:
    put unformatted "P;":U tt-sylk.val {&lf}.
  end.
  /* BNF: "P;E" <xl-font> ";M200" [ ";S" ["B"] ["I"] ]
   * "P;E" <xl-font>    Excel font.  Count from zero.
   * ";M200"            ???
   * ";S" ["B"] ["I"]   Style [Bold] [Italic]
   */
  for each tt-sylk where tt-sylk.typ = {&sylk-fil-font} no-lock by tt-sylk.seq:
    put unformatted "P;":U tt-sylk.val {&lf}.
  end.
  /* BNF: "F;" <xl-def-fmt> ";D" <fmt> <dig> <ali> <wid> ";S" <xl-def-fnt> ";M240"
   * "P;" <xl-def-fmt>   Default format number.
   * ";D" <fmt> <dig> <ali> <wid>   Default formatting, digits, alignment, and width
   *                                "G0G8" = General Format, 0 digits after the decimal,
   *                                         General (textleft, numbersright) alignment,
   *                                         8 chars wide
   * ";S" <xl-def-fnt>              Default font to use by font number
   * ";M240"                        ???
   */
  put unformatted "F;":U SylkValue({&sylk-fil-picture}, "") ";DG0G8;S":U SylkValue({&sylk-fil-font}, {&sylk-xls-body}) ";M240":U {&lf}.
  /* BNF: "B" ";Y" <num-row> ";X" <num-col> ";D0 0 1 3"
   * "B"              Boundry conditions
   * ";Y" <num-row>   Number of rows? 1 seems to work OK
   * ";X" <num-col>   Number of columns
   * ";D0 0 1 3"      ???
   */
  put unformatted "B;Y1;X":U fmtcol ";D0 0 1 3":U {&lf}.
  /* BNF: "O" ";L" ";D;V0;K47;G100 0.001"
   * "O"                      Options
   * ";L"                     Use A1 mode references (R1C1 always used in SYLK file expressions).
   * ";D;V0;K47;G100 0.001"   ???
   */
  put unformatted "O;L;D;V0;K47;G100 0.001":U {&lf}.
  /* BNF: "F;W" <begin> <end> <wid>
   * "F;W" <begin> <end> <wid>   Column width define columns <begin> to <end> as <wid> wide
   */
  for each tt-sylk where tt-sylk.typ = {&sylk-fil-c-wid} no-lock by tt-sylk.seq:
    put unformatted "F;":U tt-sylk.val {&lf}.
  end.
  /* BNF: "F;P" <xl-fmt> ";F" <fmt> <dig> <ali> ";C" <col-num>
   * "F;P" <xl-fmt>           Use format by number
   * ";F" <fmt> <dig> <ali>   formatting, digits, and alignment
   * ";C" <col-num>           Apply to column (A = 1)
   */
  for each tt-sylk
    where tt-sylk.typ = {&sylk-fil-c-fmt} no-lock by tt-sylk.seq:
    put unformatted "F;":U tt-sylk.val {&lf}.
  end.
  /* BNF: "F;F" <fmt> <dig> <ali> ";SDM" <xl-font> ";Y" <row-num> ";X" <col-num>
   *      "C;K" <q-value>
   * "F;F" <fmt> <dig> <ali>   formatting, digits, and alignment
   *                           alignment of ("L" | "R")=(Left | Right)
   * ";SDM" <xl-font>          Font to use
   * ";Y" <row-num>            Cell location
   * ";X" <col-num>            Cell location
   * "C;K" <q-value>           Cell value (quoted for text)
   */
  for each tt-sylk
    where tt-sylk.typ = {&sylk-fil-c-title} no-lock by tt-sylk.seq:
    put unformatted tt-sylk.val {&lf}.
  end.

  run tt-init-query.

  assign
    iRecordCounter = 0.

  repeat: /* process table */
    run tt-next-record.
    /* Check for a subtotal break */
    if (cBufComp <> "" and cSubBreak <> "") or (lEof and cSubTotal <> "") then
    do:
      /* Find break level */
      brk-new = num-entries(cSubBreak).
      if lEof then
        brk-new = 0.
      else
      do cnt = 1 to num-entries(cBufComp):
        find first tt-subtot where tt-subtot.bk-name = entry(cnt, cBufComp)
                                   no-lock no-error.
        if available(tt-subtot) then
          if tt-subtot.lvl < brk-new then brk-new = tt-subtot.lvl.
      end.
      if brk-new > brk-old then
        brk-old = brk-new.
      do brk-cur = brk-old to brk-new by -1:
        /* Print subtotals */
        cBreakKey  = "".
        find first tt-labels where tt-labels.bk-name = entry(brk-cur, cSubBreak)
                                   no-lock no-error.
        if available(tt-labels) then cBreakKey = tt-labels.st-lab + " " + tt-labels.st-brk.
        for each tt-subtot where tt-subtot.lvl = brk-cur no-lock by tt-subtot.st-pos:
          case tt-subtot.tot-typ:
            when {&tot-as-int} then cTotalValue = string(tt-subtot.tot-int).
            when {&tot-as-dec} then cTotalValue = string(tt-subtot.tot-dec).
            when {&tot-as-log} then cTotalValue = string(tt-subtot.tot-log).
            when {&tot-as-cnt} then cTotalValue = string(tt-subtot.tot-cnt).
          end case.
          put unformatted "C;Y":U string(iRecordCounter + 2) ";X":U string(tt-subtot.st-pos) ";K":U cTotalValue {&lf}.
        end.
        if brk-cur > 0 then cTotalValue = cBreakKey.
                       else cTotalValue = "Grand Total":U.
        put unformatted "C;Y":U string(iRecordCounter + 2) ";X":U string(num-entries(cPrevRec, {&mk})) ";K":U {&qt} cTotalValue {&qt} {&lf}.
        iRecordCounter = iRecordCounter + 1.
      end.
      /* Reset subtotals */
      for each tt-subtot where tt-subtot.lvl >= brk-new exclusive-lock:
        assign tt-subtot.tot-int = 0
               tt-subtot.tot-dec = 0.0
               tt-subtot.tot-log = 0
               tt-subtot.tot-cnt = 0.
      end.
      if lEof then
      do:
        repeat:
          brk-new = brk-new - 1.
          if brk-new < 0 then leave.
          /* Print subtotals */
          for each tt-subtot where tt-subtot.lvl = brk-new no-lock by tt-subtot.st-pos:
            case tt-subtot.tot-typ:
              when {&tot-as-int} then cTotalValue = string(tt-subtot.tot-int).
              when {&tot-as-dec} then cTotalValue = string(tt-subtot.tot-dec).
              when {&tot-as-log} then cTotalValue = string(tt-subtot.tot-log).
              when {&tot-as-cnt} then cTotalValue = string(tt-subtot.tot-cnt).
            end case.
            put unformatted "C;Y":U string(iRecordCounter + 2) ";X":U string(tt-subtot.st-pos) ";K":U cTotalValue {&lf}.
          end.
          iRecordCounter = iRecordCounter + 1.
        end.
      end.
      if brk-new > 0 or (brk-old > 0 and cSubBreak <> "") then
        iRecordCounter = iRecordCounter + 1.
    end.
    if lEof then leave.
    assign
      fmtcol    = 0
      cThisRec  = ""
      iPrnField = 0
      .
    repeat: /* process record */
      run tt-next-field.
      if lEol then leave.
      if FieldRequired(hwf) then
      do:
        iPrnField = iPrnField + 1.
        str = cFieldStr.
        if cFieldBuf = ? then str = "?":U.
        /* Record this value for comparisons */
        cThisRec = cThisRec + str + {&mk}.
        /* Save print position for subtotals */
        if iRecordCounter = 0 and cSubTotal <> "" then
        do:
          for each tt-subtot where tt-subtot.st-name = cFieldName exclusive-lock:
            assign tt-subtot.st-pos = iPrnField.
          end.
        end.
        /* check on differences with prev record (not for array fields) */
        if iRecordCounter > 0
          and can-do(cFirstOfList, cFieldName)
          and str = entry(iSeqField, cPrevRec, {&mk}) then
          str = "".

        fmtcol = fmtcol + 1.
        /* BNF: "C;Y" <row-num> ";X" <col-num> ";K" <q-value>
         * "C;Y" <row-num>   Cell location
         * ";X" <col-num>    Cell location
         * ";K" <q-value>    Cell value (quoted for text)
         */
        put unformatted "C;Y":U string(iRecordCounter + 2) ";X":U string(fmtcol) ";K":U.
        fmttmp = substring(fmtkey, iSeqField, 1, {&sChar}).
        if cFieldBuf = ? then put unformatted {&qt} "?":U {&qt}.
        else
        do:
          if fmttmp = {&SylkFormat-string} or fmttmp = {&SylkFormat-logical} then put unformatted {&qt}.
          case fmttmp:
            when {&SylkFormat-date}    then
            do:
              if date(cFieldStr) >= 1/1/1900 then
                put unformatted integer(date(cFieldStr) - 1/1/1900) + 2.
              else
                put unformatted {&qt} cFieldStr {&qt}.
            end.
            when {&SylkFormat-time}    then put unformatted integer(cFieldBuf) / 24 / 60 / 60.
            when {&SylkFormat-logical} then put unformatted right-trim(cFieldStr).
            otherwise                       put unformatted sylk-str(right-trim(cFieldBuf)).
          end case.
          if fmttmp = {&SylkFormat-string} or fmttmp = {&SylkFormat-logical} then put unformatted {&qt}.
        end.
        put unformatted {&lf}.
      end. /* Field required in output */
      else
        cThisRec = cThisRec + {&mk}.
    end. /* process record */

    /* remember 'prev' record */
    cPrevRec = cThisRec.
    iRecordCounter = iRecordCounter + 1.
  end.

  run tt-end-query.
  /* BNF: "E"
   * "E"   End of file
   */
  put unformatted "E":U {&lf}.
  output close.
end procedure. /* slk_format */


/* Name : xml_format
 * Desc : Write in XML style.
 */
procedure xml_format:
  define variable hDoc                 as handle     no-undo.
  define variable hRoot                as handle     no-undo.
  define variable hRow                 as handle     no-undo.
  define variable hText                as handle     no-undo.
  define variable hBuf                 as handle     no-undo.
  define variable hDBFld               as handle     no-undo.
  define variable iKeyField            as integer    no-undo.
  define variable cKeyField            as character  no-undo.
  define variable hKeyField            as handle     no-undo.
  define variable cNodeName            as character  no-undo.

  run tt-init-query.

  create x-document hDoc.
  create x-noderef hRoot.
  create x-noderef hRow.
  create x-noderef hField.
  create x-noderef hText.

  /* set up a root node */
  file-info:file-name = cOutputFile.
  if lAppend and file-info:full-pathname ne ? then do:
    hDoc:load("file":U, cOutputFile, false).
    /* find root node */
    hDoc:get-document-element(hRoot).
  end.
  else do:
    hDoc:create-node(hRoot, "RecordSet":U, "element":U).
    hDoc:append-child(hRoot).
  end.

  repeat:
    hQuery:get-next().
    if hQuery:query-off-end then leave.

    hDoc:create-node(hRow, btt:name, "element":U). /* create a row node */
    hRoot:append-child(hRow). /*  put the row in the tree */

    /* Filter out keyfields */
    do iKeyField = 1 to num-entries(cKeyFieldList):
      assign
        cKeyField  = entry(iKeyField, cKeyFieldList)
        hKeyField  = btt:buffer-field(cKeyField)

      /* Filter out chars XML does not like */
      cKeyField = getXMLFriendlyString( cKeyField ).

      if valid-handle(hKeyField) then
        hRow:set-attribute(cKeyField, string(hKeyField:string-value)).
    end.

    do wf = 1 to btt:num-fields:
      hwf = btt:buffer-field(wf).

      /* Filter out chars XML does not like */
      cNodeName = getXMLFriendlyString( hwf:name ).

      if FieldRequired(hwf)
        and not can-do(cKeyFieldList, hwf:name) then
      do:
        if hwf:extent = 0 then
        do:
          /* create a tag with the field name */
          hDoc:create-node(hField, cNodeName, "element":U).

          /* put the new field as next child of row */
          hRow:append-child(hField).

          /* add a node to hold field value */
          hDoc:create-node(hText, "", "text":U).

          /* attach the text to the field */
          hField:append-child(hText).
          hText:node-value = string(hwf:string-value).
        end.
        else
        do nx = 1 to hwf:extent:
          /* create a tag with the field name */
          hDoc:create-node(hField, substitute("&1_&2":U, cNodeName, nx), "element":U).

          /* put the new field as next child of row */
          hRow:append-child(hField).

          /* add a node to hold field value */
          hDoc:create-node(hText, "", "text":U).

          /* attach the text to the field */
          hField:append-child(hText).
          hText:node-value = string(hwf:string-value(nx)).
        end. /* extent > 1 */
      end. /* field required in output */
    end. /* for each field */
  end. /* repeat */

  hQuery:query-close().
  delete object hQuery.

  /* write the XML node tree to an xml file */
  hDoc:save("file":U, cOutputFile).
  delete object hDoc.
  delete object hRoot.
  delete object hRow.
  delete object hField.
  delete object hText.
end procedure. /* xml_format */


/* Name : fatalError
 * Desc : Exit due to fatal error.
 */
procedure fatalError.
  define input parameter cErrorMsg as character no-undo.

  if cOutputFile <> "" then
  do:
    output to value(cOutputFile).
    put unformatted cErrorMsg {&lf}.
    output close.
  end.

  message cErrorMsg
    view-as alert-box information buttons ok.

end procedure. /* fatalError */
